#include <PalmOS.h>
#include <SysEvtMgr.h>
#include "resource.h"

#define expired() ((prefs.evalStart != 'PAID') && (TimGetSeconds() - prefs.evalStart > 86400 * 30))
#define ourMinVersion 2

#define ScoreLength 10

/* Internal Structures */
typedef struct {
	UInt32 evalStart;
	UInt16 highscores[ScoreLength];
} MyPreferenceType;

static MyPreferenceType prefs;

/*
typedef struct {
	// Byte replaceme;
} MyAppInfoType;

static MyAppInfoType appInfo;
*/

#define ObjectQty 12

typedef enum { VOID = 0, ADD, SUB, MUL, DIV, MOD } Operator;

typedef struct {
	UInt8 left;
	UInt8 right;
	Operator operator;
	Coord posX;
} Object;

Object obj[ObjectQty];

/* Internal Constants */
#define appVersionNum     0x01;
#define appPrefID         0x00
#define appPrefVersionNum 0x01
#define appFileCreator   'HkZn'

UInt32 timeToMove = 0;

#define TicksPerRound (SysTicksPerSecond() * 2)
#define handLengthMax 10
Char hand[handLengthMax];
UInt8 handLength;
typedef enum {NotRunning, Running} State;
State state;
UInt16 missQty = 0;
UInt16 eraseQty = 0;

static void gameOver();

static UInt16 RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags) {
	UInt32 romVersion;
	// See if we're on in minimum required version of the ROM or later.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion) {
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
				(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
			FrmAlert (SystemIncompatible);
			// Pilot 1.0 will continuously relaunch this app unless we switch to
			// another safe one.
			if (romVersion < sysMakeROMVersion(2,0,0,sysROMStageRelease,0)) {
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
			}
		}
		return sysErrRomIncompatible;
	}
	return 0;
}

static void *GetObjectPtr(UInt16 objectID) {
	FormType *frmP;
	frmP = FrmGetActiveForm();
	return (FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID)));
}

/* max of number of digits of object:
                         max(log(left) + log(right)) */
#define answerLengthMax 5

static void createObject(Object *obj) {
    int operLevel = eraseQty%100;
    int range = eraseQty/100 + 1;
    UInt8 i;
	obj->posX = SysRandom(0) % 110 + 10;
	obj->operator =
    	 (operLevel<20)?ADD:
    	 (operLevel<40)?SUB:
    	 (operLevel<60)?SysRandom(0)%2+1:
    	 (operLevel<80)?SysRandom(0)%2*2+1:
                    	SysRandom(0)%3+1;
	obj->left = SysRandom(0) % (9*range) + 11;
	obj->right = SysRandom(0) % (6*range) + 3;
	if(obj->operator == SUB && (obj->left < obj->right)) {
	    i = obj->left; obj->left = obj->right; obj->right = i;
	}
}

static UInt16 evalObject(Object *obj) {
	return obj->operator == ADD ? obj->left + obj->right :
		   obj->operator == SUB ? obj->left - obj->right :
		   obj->operator == MUL ? obj->left * obj->right :
		   obj->operator == DIV ? obj->left / obj->right :
		   obj->operator == MOD ? obj->left % obj->right :
		   VOID;
}

static void hideObject() {
	RectangleType rect;
	rect.topLeft.x = 0; rect.topLeft.y = 15;
	rect.extent.x = 160; rect.extent.y = 132;
	WinEraseRectangle(&rect,0);
}

static void drawObject() {
	Char buffer[8];
	int index;
	int i;
	hideObject();
	for(i=0; i<ObjectQty; i++) {
		if(obj[i].operator==VOID) continue;
		index = 0;
		StrIToA(buffer, obj[i].left);
		index += (obj[i].left < 10 ? 1 : 2);
		buffer[index++]=32;
		buffer[index++]="_+-*/%"[obj[i].operator];
		buffer[index++]=32;
		StrIToA(buffer+index, obj[i].right);
		index += (obj[i].right < 10 ? 1 : 2);
		buffer[index]='\0';
		WinDrawChars(buffer,index,obj[i].posX,15 + i*11);
	}
}

#define missQtyOfGameOver 10

/** @return if gameOver:false */
static Boolean MoveObject() {
	int i;
	char buffer[3];
	if(obj[ObjectQty-1].operator!=VOID) {
    	for(i=0; i<10; i++) {
    	    WinDrawLine(obj[ObjectQty-1].posX+16,147,obj[ObjectQty-1].posX+(SysRandom(0)&31),147+(SysRandom(0)&7));
    	}
    	WinEraseLine(20+missQty*2,150,20+missQty*2,159);
    	if(missQty++ >= missQtyOfGameOver) {
    	    gameOver();
    	    return false;
    	}
	}
	for(i = ObjectQty - 2 ; i >= 0; i--) {
		obj[i+1] = obj[i];
	}
	createObject(&obj[0]);
	drawObject();
	return true;
}

/* add digit to hands
 * if digit entered, add it to hands and draw them
 */
static void addHand(WChar entered) {
	int i;
	if(handLength < handLengthMax) {
		hand[handLength++] = entered;
	} else {
	    for(i=0;i<handLength-1;i++) hand[i]=hand[i+1];
	    hand[handLength-1] = entered;
    }
	WinDrawChars(hand,handLength, 160-5*handLength,150);
}

static Boolean delHand() {
	RectangleType rect;
    if(handLength>0) {
        handLength-=1;
		rect.topLeft.x = 110; rect.topLeft.y = 150;
		rect.extent.x = 50; rect.extent.y = 10;
		WinEraseRectangle(&rect,0);
    	WinDrawChars(hand,handLength, 160-5*handLength,150);
    	return true;
	} else {
    	return false;
	}
}

/* shoot */
static void shoot() {
/* Logic implemented as following:
		   do
			[123456] start matching with 3digits(=MAX)
				offset ->
			len 123 234 345 456 567
				12 23 34 45 56 67 <- matched with "45", "56"
			[1237] second tarn; start matching with 3digits
				123 247
				12 23 37 <- matched two object with "37"
			[12] third tarn; start matching with 2digits
				12
				1 2 <- matched with "2"
			[1]
				1 <- not matched; exit evaluating
		*/
	Boolean matched,valmatched;
	RectangleType rect;
	UInt8 bullQty = 0; // PENDING: bonus?
	UInt8 tempHandLength;
	UInt16 val;
	int len, offset, i, j;
	Boolean matchedFlag[handLength];
    Char buffer[maxStrIToALen];
	do {
		matched=false;
		for(i=0; i<handLength; matchedFlag[i++]=false);
		// ]錅Zo(őanswerLengthMax)
		tempHandLength = handLength;
		if(tempHandLength > answerLengthMax ) tempHandLength = answerLengthMax;
		//ق}b`OĂ
		for(len = tempHandLength; len>0; len--) {
			// with sliding offset;
			for(offset=0; offset<=handLength-len; offset++) {
				valmatched=false;
				if(len!=1 && hand[offset]==chrDigitZero) continue; // Ń[Ŏn܂ꍇ𖳎
				// calc for matching;
				for(val=0, i=0; i<len; i++) {
					val = val * 10 + hand[offset+i] - chrDigitZero;
				}
				// lIuWFNǧvZlƔrĂ
				for(i=0; i<ObjectQty; i++) {
					if(obj[i].operator != VOID
							&& val == evalObject(&obj[i])) { // v!
						valmatched=true;
						// IuWFNg
						WinDrawLine(0,20+11*i,160,20+11*i);
						obj[i].operator = VOID;
						eraseQty++;
					}
				}
				// }b`DɃ}[NĂ
				if(valmatched) {
    				matched=true;
					for(i=0;i<len;i++) {
						matchedFlag[offset+i]=true;
                    	WinDrawLine(160-5*(handLength-offset-i),159,165-5*(handLength-offset-i),159);
					}
				}
			}
			if(matched) {
				// D}b`菜
				for(i=0,j=0; i<handLength; i++) {
					if(!matchedFlag[i]) hand[j++]=hand[i];
				}
				handLength=j;
				rect.topLeft.x = 110; rect.topLeft.y = 150;
				rect.extent.x = 50; rect.extent.y = 10;
				// ̉sAǂoX悭Ȃ
				// ƎvǂȂ񂩂
        		SysTaskDelay(20);
        		timeToMove += 20;
				WinEraseRectangle(&rect,0);
            	WinDrawChars(hand,handLength, 160-5*handLength,150);
				// }b`炻菭Ȃ̓`FbNȂƂɂ
				//  A_ɓꂽق
				break;
			}
		}
		drawObject();
        StrIToA(buffer,eraseQty);
        WinDrawChars(buffer,StrLen(buffer),0,150);
				// A_
	} while(matched);
    if(expired() && eraseQty >= demoQty) {
        if(eraseQty%5==0) FrmAlert(PleaseRegister);
    }
}

static void gameOver() {
    int i;
    UInt16 scoretemp;
    Char buffer[maxStrIToALen+14];
    ControlType *button;
    RectangleType rect;
    rect.topLeft.x = 20; rect.topLeft.y = 30;
    rect.extent.x = 120; rect.extent.y = 100;
    WinEraseRectangle(&rect,0);
    WinDrawChars("G A M E  O V E R",16,50,38);
    StrCopy(buffer,"Your Score is ");
    StrIToA(buffer+14,eraseQty);
    WinDrawChars(buffer,StrLen(buffer),50,50);
    CtlDrawControl((ControlType*)GetObjectPtr(StartButton));
    for(i=0; i<ScoreLength; i++) {
        if(prefs.highscores[i] <= eraseQty) {
            scoretemp = prefs.highscores[i];
            prefs.highscores[i] = eraseQty;
            eraseQty = scoretemp;
        }
    }


    WinDrawChars("Best Scores:",12,50,92);
    StrIToA(buffer,prefs.highscores[0]); WinDrawChars(buffer,StrLen(buffer),105, 92);
    StrIToA(buffer,prefs.highscores[1]); WinDrawChars(buffer,StrLen(buffer),105,102);
    StrIToA(buffer,prefs.highscores[2]); WinDrawChars(buffer,StrLen(buffer),105,112);
    state=NotRunning;
    timeToMove=0;
}

static void gameStart() {
    int i;
    RectangleType rect;
	while(delHand());
	for(i=0; i<ObjectQty; obj[i++].operator=VOID);
	eraseQty=0;
	missQty=0;
	rect.topLeft.x = 0; rect.topLeft.y = 15;
	rect.extent.x = 160; rect.extent.y = 144;
    WinEraseRectangle(&rect,0);
    WinDrawLine(0,147,159,147);
	for(i=0;i<missQtyOfGameOver;i++)
	    WinDrawLine(20+i*2,150,20+i*2,159);
	WinDrawChars("answers -->",11,60,150);
	timeToMove = TimGetTicks();
	state = Running;
}

static void pause() {
    if(state == Running) {
        timeToMove = 0;
    }
}

static void resume() {
    if(state == Running) {
        drawObject();
        timeToMove = TimGetTicks() + TicksPerRound;
    }
}

static void MainFormInit() {
}

/*
static Boolean blipState;

static void blip() {
    RectangleType rect;
    rect.topLeft.x = 0; rect.topLeft.y = 0;
    rect.extent.x = 15; rect.extent.y = 15;
    if(blipState =! blipState)
        WinDrawRectangle(&rect,0);
    else
        WinEraseRectangle(&rect,0);
}
*/

static Boolean MainFormHandleEvent(EventType *eventP) {
	FormType *frmP;
	int i;
	WChar entered;
	MenuBarType *menu;
	Boolean handled = false;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		break;
	case menuEvent:
		switch (eventP->data.menu.itemID) {
		case MenuOption:
    		pause();
    		FrmAlert(NotImplemented);
			resume();
			handled = true;
			break;
		case MenuInstruction:
    		pause();
    		FrmAlert(Instruction);
			resume();
			handled = true;
			break;
		case MenuAbout:
    		FrmPopupForm(AboutForm);
			handled = true;
			break;
		default:
			break;
		}
		break;
	case keyDownEvent:
	    if(state==Running) {
    		entered = eventP->data.keyDown.chr;
    		if(entered == chrSmall_O) entered=chrDigitZero;
    		if(entered == chrSmall_U) entered=chrDigitZero;
    		if(entered == chrSmall_I) entered=chrDigitOne;
    		if(entered == chrSmall_Z) entered=chrDigitTwo;
    		if(entered == chrSmall_B) entered=chrDigitThree;
    		if(entered == chrSmall_C) entered=chrDigitFour;
    		if(entered == chrSmall_L) entered=chrDigitFour;
    		if(entered == chrSmall_S) entered=chrDigitFive;
    		if(entered == chrSmall_G) entered=chrDigitSix;
    		if(entered == chrSmall_T) entered=chrDigitSeven;
    		if(entered == chrSmall_Y) entered=chrDigitEight;
    		if(entered == chrSmall_Q) entered=chrDigitNine;
    		if(entered >= chrDigitZero && entered <= chrDigitNine) {
    			addHand(entered);
    			handled = true;
    		} else if (entered == chrBackspace) {
    		    delHand();
    		} else if (entered == chrLineFeed
    		        || entered == chrSpace
                	|| entered == chrCarriageReturn
    		        || entered == chrVerticalTabulation
    		        || entered == chrFormFeed
    		) {
    			shoot();
    			handled = true;
    		}
		}
		break;
	case frmCloseEvent:
		break;
	case ctlSelectEvent:
	    if(eventP->data.ctlSelect.controlID != StartButton)
	        break;
        // if START tapped,
        if(state==Running) {
            // DEBUG USE:
            Char buffer[maxStrIToALen];
            StrIToA(buffer,timeToMove);
            WinDrawChars(buffer,StrLen(buffer),0,120);
            StrIToA(buffer,TimGetTicks());
            WinDrawChars(buffer,StrLen(buffer),0,130);
            break;
        }
        gameStart();
		/* no break; Do First Move */
	case nilEvent:
		if(state == Running
        		&& timeToMove != 0
        		&& timeToMove <= TimGetTicks()
        		&& !(menu = MenuGetActiveMenu())
        		 || !menu->attr.visible) {
			if(MoveObject())
    			timeToMove += TicksPerRound;
		}
		handled=true;
		break;
	default:
		break;
	}
	return handled;
}

/**
 * About Form
 */
static Boolean AboutFormHandleEvent(EventPtr eventP) {
	UInt8 x,y;
	Boolean handled = false;
	FormPtr frmP;
	switch (eventP->eType)	{
	case frmOpenEvent:
		frmP = FrmGetActiveForm();
		FrmDrawForm ( frmP);
		WinDrawChars( prefs.evalStart == 'PAID' ? "ALREADY REGISTERED" :
                            		 !expired() ? " UNDER EVALUATION " :
                            		              "   DEMO EXPIRED   "
                            		,18,35,100);
	/* for DEBUG
    rect.topLeft.x = 77; rect.topLeft.y = 67;
    rect.extent.x = 7; rect.extent.y = 8;
    WinDrawRectangle(&rect,0);
       *** */
		pause();
		break;
	case penDownEvent:
		x = eventP->screenX;
		y = eventP->screenY;
		if(77 <= x && x <= 84 && 67 <= y && y <= 75) {
            prefs.evalStart = 'PAID';
            FrmAlert(ThanksForRegister);
        }
		WinDrawChars( prefs.evalStart == 'PAID' ? "ALREADY REGISTERED" :
                            		 !expired() ? " UNDER EVALUATION " :
                            		              "   DEMO EXPIRED   "
                            		,18,35,100);
		break;
	case ctlSelectEvent:
		switch (eventP->data.ctlSelect.controlID) {
		case CloseButton:
    		if(KeyCurrentState() & keyBitPageDown) {
                prefs.evalStart = 'PAID';
                FrmAlert(ThanksForRegister);
            }
			FrmReturnToForm(MainForm);
		    resume();
			handled = true;
			break;
		default:
			break;
		}
		break;
	}
	return handled;
}

static Boolean AppHandleEvent( EventType *eventP) {
	UInt16 formId;
	FormType *frmP;
	Boolean handled = false;
	switch (eventP->eType) {
	case frmLoadEvent:
		// Load the form resource.
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
		switch (formId) {
		case MainForm:
			MainFormInit();
			FrmSetEventHandler(frmP, MainFormHandleEvent);
			handled = true;
			break;
		case AboutForm:
			FrmSetEventHandler(frmP, AboutFormHandleEvent);
			handled = true;
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
	return handled;
}

static void AppEventLoop(void) {
	UInt16 error;
	UInt32 wait;
	UInt32 currentTime;
	EventType event;
	do {
	    currentTime = TimGetTicks();
		wait = (timeToMove == 0) ? evtWaitForever :
               (timeToMove <= currentTime) ? TicksPerRound / 3 :
               (timeToMove - currentTime);
		EvtGetEvent(&event, wait);
		SysHandleEvent(&event)
		|| MenuHandleEvent(0, &event, &error)
		|| AppHandleEvent(&event)
		|| FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

/**
 * Get the current application's preferences.
 * @return Err value 0 if nothing went wrong
 */
static UInt16 AppStart(void) {
    int i;
	UInt16 prefsSize;
	DateType date;
	UInt32 today ;

	SysRandom(TimGetTicks());
	timeToMove = 0;
	state = NotRunning;

	// Read the saved preferences / saved-state information.
	prefsSize = sizeof(MyPreferenceType);
	if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, false)
        	== noPreferenceFound) {
		prefs.evalStart = REGISTERED ? 'PAID' : TimGetSeconds();
		for(i=0; i<ScoreLength; i++) prefs.highscores[i] = 0;
	}
	if(expired()) FrmAlert(PleaseRegister);
	return 0;
}

/**
 * Save the current state of the application.
 */
static void AppStop(void) {
	// Write the saved preferences / saved-state information.  This data
	// will be backed up during a HotSync.
	int i;
	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum,
		&prefs, sizeof (prefs), false);
    FrmCloseAllForms();
}

/**
 * This is the main entry point for the application.
 * @param cmd		  Word value specifying the launch code.
 * @param cmdPB		  Pointer to a structure that is associated with the launch code.
 * @param launchFlags Word value providing extra information about the launch.
 * @return Result of launch
 */
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags) {
	UInt16 error;
	switch (cmd) {
	case sysAppLaunchCmdNormalLaunch:
		error = RomVersionCompatible (ourMinVersion, launchFlags);
		if (error) return error;
		error = AppStart();
		if (error) return error;
		FrmGotoForm(MainForm);
		AppEventLoop();
		AppStop();
		break;
	default:
		break;
	}
	return 0;
}
